/**
 * Create Solid component what is base class
 * for all solid objects in game
 */ 
Crafty.c("Solid", {
  init: function() {
    var me = this;
    me.requires("2D, DOM");
  }
});

/**
 * Create CharacterComponent what is
 * base component for Player and Enemy component
 */ 
Crafty.c("CharacterComponent", {
  init: function() {
    var me = this;
    me.LEFT = 0;
    me.RIGHT = 1;
    me.direction = me.RIGHT;
    me.requires("Solid, Gravity, Collision");
    me.gravity("Ground");
    me.onHit("Portal", $.proxy(me.onPortalHit, me));
  },
  
  /**
   * When character hits portal then teleport
   * 
   * @return void
   */ 
  onPortalHit: function(portals) {
    var me = this,
        portal = portals[0].obj,
        to = Fuzed.portals[portal.to];
        
    if (to && (me.direction === me.LEFT && me.x < portal.x+portal.w/2 || (me.direction === me.RIGHT && me.x+me.w > portal.x+portal.w/2))) {
      me.x = to.x + 10;
      me.y = to.y + (to.h-me.h);
    }
  }
});

/**
 * Create EnemyBatch component for
 * performing movement of all enemies using one EnterFrame event listener
 */ 
Crafty.c("EnemyBatch", {
  init: function() {
    this.requires("Solid");
    this.attr({
      x: 0,
      y: 0,
      z: 1
    })
  }
});

/**
 * Create enemy component
 */ 
Crafty.c("EnemyComponent", {
  init: function() {
    var me = this;
    me.health = 2;
    me.direction = Crafty.math.randomInt(0, 1);
    me.requires("CharacterComponent");
    me.collision(new Crafty.polygon([8,8], [24, 8], [24, 32], [8, 32]));
    
    me.attr({
      x: Crafty.math.randomInt(128, Fuzed.world[0]-128), 
      y: 0,
      z: 100, 
      _speed: Crafty.math.randomNumber(0.5, 1.5) 
    });
    
    me.onHit("Wall", $.proxy(me.onWallHit, me));
    me.onHit("Explosion", $.proxy(me.onExplosionHit, me));
  },
  
  /**
   * This function is called on every EnterFrame event
   * @return void
   */ 
  onEnterFrame: function() {
    var me = this,
        dom = $(me._element);
    if (me.health === 0) {
      return;
    }
    
    if (me.direction === me.LEFT) {
      if (!dom.hasClass(me.name+"-move-left")) {
        dom.removeClass(me.name+"-move-right");
        dom.addClass(me.name+"-move-left");
      }
      me.x -= me._speed;
    }
    else if (me.direction === me.RIGHT) {
      if (!dom.hasClass(me.name+"-move-right")) {
        dom.removeClass(me.name+"-move-left");
        dom.addClass(me.name+"-move-right");
      }
      me.x += me._speed;
    }
    
    // If falls into hold, then throw back on stage
    if (me.y+me.h > Fuzed.world[1]) {
      me.x = Crafty.math.randomInt(128, Fuzed.world[0]-128);
      me.y = 0;
    }
    
    if (me.x+me.w/2 < 0) {
      me.x = Fuzed.world[0]-me.w;
    } else if (me.x+me.w/2 > Fuzed.world[0]) {
      me.x = 0;
    }
  },
  
  /**
   * Hurt enemy
   * 
   * @param {Integer} amount
   * @return void
   */ 
  hurt: function(amount) {
    var me = this,
        dom = $(me._element);
        
    if (me.health) {
      me.health -= amount;
      // Die
      if (me.health <= 0) {
        me.health = 0;
        dom.removeClass(me.name+"-move-left");
        dom.removeClass(me.name+"-move-right");
        dom.addClass(me.name+"-die");
        me.y -= me.h/2;
        setTimeout(function() {
          me.destroy();
        }, 300);
      }
    }
  },
  
  /**
   * On wall hit change move direction
   * 
   * return void
   */ 
  onWallHit: function(wall) {
    var me = this,
        obj = wall[0].obj;
    if (obj.x < me.x) {
      me.x += me._speed;
      me.direction = me.RIGHT;
    }
    else {
      me.x -= me._speed;
      me.direction = me.LEFT;
    }
  },
  /**
   * When enemy hits explosion it dies
   * 
   * return void
   */ 
  onExplosionHit: function() {
    this.hurt(1);
  }
});

/**
 * Create portal component
 */ 
Crafty.c("Portal", {
  init: function() {
    this.requires("Solid, portal_sprite");
  }
});

/**
 * Create coin component
 */ 
Crafty.c("Coin", {
  init: function() {
    var me = this;
    me.requires("Solid, coin_gold, Collision");
    me.collision(new Crafty.circle(8, 8, 8));
  }
});

/**
 * Create mine component
 */ 
Crafty.c("Mine", {
  init: function() {
    var me = this;
    me.exploded = false;
    me.requires("Solid, mine_floor, Gravity, Collision");
    me.collision(new Crafty.polygon([1, 10], [15, 10], [15, 16], [1, 16]));
    me.gravity("Ground");
    
    // Init at random coordinates
    me.attr({
      x: Crafty.math.randomInt(128, Fuzed.world[0]-128), 
      y: Crafty.math.randomInt(0, Fuzed.world[1]-16)
    });
    
    // Randomizes. JavaScript has pretty bad random algorithm
    Crafty.math.randomInt(1, 20000);
  },
  
  /**
   * Triggers explosion animation for mine and
   * destroies mine instance
   * 
   * @return void
   */ 
  explode: function() {
    var me = this;
    setTimeout(function() {
      Crafty.e("Explosion").attr({
        x: me.x - 24,
        y: me.y - 48
      });
      me.exploded = true;
      me.destroy();
    }, 200);
  }
});

/**
 * Create bomb component
 */ 
Crafty.c("Bomb", {
  init: function() {
    var me = this;
    me.exploded = false;
    me.requires("Solid, bomb_floor, Gravity");
    me.gravity("Ground");
    // Explosion is triggered 1 second after deployment
    setTimeout($.proxy(me.explode,me), 1000);
  },
  
  /**
   * Triggers explosion animation for bomb and
   * destroies bomb instance
   * 
   * @return void
   */ 
  explode: function() {
    var me = this;
    Crafty.e("Explosion").attr({
      x: me.x - 24,
      y: me.y - 48
    });
    me.exploded = true;
    me.destroy();
  }
});

/**
 * Define Explosion component
 * Explosion animation is started when instance is created
 */ 
Crafty.c("Explosion", {
  init: function() {
    var me = this;
    me.requires("Solid, explosion_sprite, Collision");
    me.collision(new Crafty.polygon([16, 32], [48, 32], [48, 64], [16, 48]));
    // Destory explosion component after 1.2 seconds
    setTimeout($.proxy(me.destroy, me), 1200);
  }
});

/**
 * Define player component
 */ 
Crafty.c("Player", {
  init: function() {
    var me = this;
    me.health = 100;
    me.coins = 0;
    me.bombs = 10;
    me.requires("CharacterComponent, Keyboard, gripe");
    me.attr({
      x: 64, 
      y: Fuzed.world[1]-48, 
      z: 100,
      _speed: 3,
      _jump: 3
    });
    
    // Define collision polygon for player
    me.collision(new Crafty.polygon([8,8], [24, 8], [24, 32], [8, 32]));
    me.bind("EnterFrame", $.proxy(me.onEnterFrame, me));
    // Wall colision listener
    me.onHit("Wall", $.proxy(me.onWallHit, me));
    
    // One collision listener for all solid object
    // We would loose performace if we would have
    // collision listener for every object class
    me.onHit("Solid", function(comps) {
      var first = comps[0].obj;
      if (first.has("Coin")) {
        me.onCoinHit(comps);
      }
      else if (first.has("EnemyComponent")) {
        me.onEnemyHit(comps);
      }
      else if (first.has("Mine")) {
        me.onMineHit(comps);
      }
      else if (first.has("Explosion")) {
        me.onExplosionHit(comps);
      }
    });
    
    // Define animation classes
    me.animateClasses = [
      "gripe-stand-right",
      "gripe-stand-left",
      "gripe-move-right",
      "gripe-move-left",
      "gripe-jump-right",
      "gripe-jump-left"
    ];
  },
  
  /**
   * Callback for EnterFrame callback what is execute at every frame
   * @return void
   */
  onEnterFrame: function() {
    var me = this;
    
    if (me._up) {
      me.y -= me._jump;
      me._falling = true;
    }

    if (navigator.accelerometer) {
      navigator.accelerometer.getCurrentAcceleration($.proxy(me.moveAccelometer, me), function() {
        if (!me._accelerometerfailed) {
          alert("No accelerometer data!");
          me._accelerometerfailed = true;
        }
      });
    }
    else {
      me.moveKeyboard();
    }
    
    // Fall into hole
    if (me.y+me.h >= Fuzed.world[1]) {
      me.hurt(100);
    }
    // Enable level walktrough
    if (me.x+me.w/2 < 0) {
      me.x = Fuzed.world[0]-me.w;
    } else if (me.x+me.w/2 > Fuzed.world[0]) {
      me.x = 0;
    }
  },
  
  /**
   * Initializes mobile movement
   * @return void
   */
  moveAccelometer: function(acc) {
    var me = this;
    var iphone = /iphone/.test(navigator.userAgent.toLowerCase());

    if (acc.y > 0 && acc.y >= 2) {
      me[iphone ? "moveLeftAction" : "moveRightAction"]();
    }
    else if (acc.y < 0 && Math.abs(acc.y) >= 2) {
      me[iphone ? "moveRightAction" : "moveLeftAction"]();
    }
    else {
      me.standAction();
    }
  },
  
  /**
   * Initializes keyboard listeners
   * @return void
   */ 
  moveKeyboard: function() {
    var me = this;
    if (me.isDown("UP_ARROW")) {
      me.jumpAction();
    }
    
    if (me.bombs > 0 && me.isDown(Crafty.keys.SPACE)) {
      me.plantBombAction();
    }
    
    if (me.isDown(Crafty.keys.RIGHT_ARROW)) {
      me.moveRightAction();
    }
    else if (me.isDown(Crafty.keys.LEFT_ARROW)) {
      me.moveLeftAction();
    }
    else {
      me.standAction();
    }
  },
  
  /**
   * Add animation class for player.
   * 
   * @param {String} cls
   * @return void
   */ 
  addAnimClass: function(cls) {
    var me = this,
        dom = $(me._element);
    if (!dom.hasClass(cls)) {
      dom.removeClass(me.animateClasses.join(" "));
      dom.addClass(cls);
    }
  },
  
  /**
   * Stand action
   * 
   * @return void
   */ 
  standAction: function() {
    var me = this,
        standClass = "gripe-stand-" + (me.direction === me.LEFT ? "left" : "right");
    me.addAnimClass(standClass);
  },
  
  /**
   * Jump action
   * 
   * @return void
   */ 
  jumpAction: function() {
    this._up = true;
  },
  
  /**
   * Move right action
   * 
   * @return void
   */ 
  moveRightAction: function() {
    var me = this;
    if (me._up) {
      me.addAnimClass("gripe-jump-right");
    }
    else {
      me.addAnimClass("gripe-move-right");
    }
    me.x += me._speed;
    me.direction = me.RIGHT;
  },
  
  /**
   * Move left action
   * 
   * @return void
   */ 
  moveLeftAction: function() {
    var me = this;
        
    if (me._up) {
        me.addAnimClass("gripe-jump-left");
      }
      else {
        me.addAnimClass("gripe-move-left");
      }
      me.x -= me._speed;
      me.direction = me.LEFT;
  },
  
  /**
   * Action for planting the bomb
   * 
   * @return void
   */ 
  plantBombAction: function() {
    var me = this,
        dom = $("#bombs .value");
    
    if (!me._plantingBomb) {
      me.bombs -= 1;
      dom.html(+dom.html()-1);
      // But bomb on ground
      Crafty.e("Bomb")
      .attr({
        x: me.x + (me.direction === me.LEFT ? 0 : 16),
        y: me.y + (me.h-16),
        z: 12000
      });
      me._plantingBomb = true;
      // Sleep before allowing next bomb
      setTimeout(function() {
        me._plantingBomb = false;
      }, 500);
    }
  },
  
  /**
   * Hurt player. If player health drops below zero "gameOver"
   * event is triggered.
   * 
   * @param {Integer} amount
   * @return void
   */
  hurt: function(amount) {
    var me = this;
    if (me.health) {
      // If we have mobile then vibrate on hit
      navigator.notification && navigator.notification.vibrate(1000);
      me.health -= amount;
      if (me.health <= 0) {
        me.health = 0;
        me.trigger("gameOver");
        me.destroy();
      }
      //  Update statusbar
      $("#health .value").html(me.health);
    }
  },
  
  /**
   * Callback for wallhit
   * 
   * @param {Array} wall
   * @return void
   */
  onWallHit: function(wall) {
    var me = this,
        obj = wall[0].obj;
    if (obj.x < me.x) {
      me.x += me._speed;
    }
    else {
      me.x -= me._speed;
    }
  },
  
  /**
   * Callback for coin hit
   * 
   * @param {Array} coin
   * @return void
   */
  onCoinHit: function(coin) {
    var me = this;
    $.each(coin, function(i, c) {
      me.coins++;
      c.obj.destroy();
    });

    // Update statusbar
    $("#coins .value").html(+$("#coins .value").html()+coin.length);
  },
  
  /**
   * Callback for explosion collision
   * 
   * @param {Array} expl
   * @return void
   */
  onExplosionHit: function(expl) {
    var me = this,
        e = expl[0];
    
    if (!me.explosionsSurvived) {
      me.explosionsSurvived = {};
    }
    
    if (!me.explosionsSurvived[e.obj[0]]) {
      me.explosionsSurvived[e.obj[0]] = true;
      me.hurt(10);
    }
  },
  
  /**
   * Callback for enemy collision
   * 
   * @param {Array} enemy
   * @return void
   */
  onEnemyHit: function(enemy) {
    this.hurt(2);
  },
  
  /**
   * Callback what is triggered when player
   * collides with mine object
   * 
   * @param {Array} mines
   * @return void
   */
  onMineHit: function (mines) {
    var mine = mines[0].obj;
    if (!mine.exploded) {
      mine.explode();
    }
  }
});